#include "swMathExt.h"
#include <limits>
#include <cmath>
#include <cassert>

#if defined(max)
#	undef max
#endif


//-----------------------------------------------------------------------------------------
//

double logb(double base, double x)
{
	double dTol = swMathConst::EPS;
	if (base <= 0.0 || (-dTol <= base-1.0 && base-1.0 <= dTol)) {
		assert(false);
		return 0.0;
	}
	if (x <= 0.0) {
		assert(false);
		return 0.0;
	}

	return log(x) / log(base);
}


double asinh(double x)
{  return log(x + sqrt(x*x + 1.0));  }

double acosh(double x)
{
	if (x < 1.0) {
		//  when x < 1.0, a solution is a conmplex number
		assert(false);
		return 0.0;
	}

	return log(x + sqrt(x*x - 1.0));
	//return log(x - sqrt(x*x - 1.0));
}

double atanh(double x)
{
	if (x <= -1.0 || x >= 1.0) {
		//  when x <= -1.0 || x >= 1.0, a solution is a conmplex number
		assert(false);
		return 0.0;
	}

	return log(sqrt((1.0 + x) / (1.0 - x)));
}


//-----------------------------------------------------------------------------------------
//  struct swMathExt

bool swMathExt::is_integer(double x, double dTol/* = swMathConst::EPS*/)
{

	double dInteger, dFraction;
	dFraction = modf(x, &dInteger);
	return -dTol <= dFraction && dFraction <= dTol;
	//return std::numeric_limits<T>::is_integer;
}

bool swMathExt::is_real(double x, double dTol/* = swMathConst::EPS*/)
{
	double dInteger, dFraction;
	dFraction = modf(x, &dInteger);
	return dFraction < -dTol || dFraction > dTol;
	//return !std::numeric_limits<T>::is_integer;
}

double swMathExt::mantissa(double x, double* pow/* = 0L*/)
//  value = mentissa * 10^power;
{
	double dSign = sign(x);
	x = fabs(x);

	if (x >= 1.0)
		//  123.45 = 1.2345 * 10^2
		for (unsigned long i=0ul ; i<=std::numeric_limits<unsigned long>::max() ; ++i) {
			if (x < 10.0) {
				if (pow) *pow = double(i);
				return x * dSign;
			}
			x /= 10.0;
		}
	else
		//  0.006789 = 6.789 * 10^-3
		for (unsigned long i=1ul ; i<=std::numeric_limits<unsigned long>::max() ; ++i) {
			x *= 10.0;
			if (x >= 1.0) {
				if (pow) *pow = -double(i);
				return x * dSign;
			}
		}
			
	if (pow) *pow = 0.0;
	return 0.0;
}

double swMathExt::round(double x)
//  round(-2.51) = -3, round(-2.50) = -2, round(-2.49) = -2
//  round(2.49) = 2, round(2.50) = 3, round(2.51) = 3
{
	//return floor(x + 0.5);
	return x > 0.0 ? int(x + 0.5) : int(x - int(x) + 1.5) + int(x) - 1;
}

double swMathExt::round(double x, int decimalPlace)
//  round(12340.98, 3) is 12000.00
//  round(0.003456, -4) is 0.003500
{
/*
	if (fabs(x) >= 1.0) {
		x *= pow(10.0, double(-decimalPlace));
		x = x >= 0.0 ? floor(x + 0.5) : ceil(x - 0.5);
		x *= pow(10.0, double(decimalPlace));
	}
	else {
		x *= pow(10.0, double(-decimalPlace - 1));
		x = x >= 0.0 ? floor(x + 0.5) : ceil(x - 0.5);
		x *= pow(10.0, double(decimalPlace + 1));
	}
	
	return x;
*/
	x /= pow(10.0, decimalPlace);
	x = round(x);
	return x * pow(10.0, decimalPlace);
}

unsigned long swMathExt::gcd(unsigned long lhs, unsigned long rhs)
//  GCD  :  Greatest Common Divisor
//  use Euclidean Algorithm
//  if return value is one(1), GCD does not exist;
{
/*
	if (lhs == rhs) return lhs;

	if (lhs < rhs)
	if (!(rhs %= lhs)) return lhs;
*/    
	while (1) {
		if (!(lhs %= rhs)) return rhs;
		if (!(rhs %= lhs)) return lhs;
	} 
}

double swMathExt::gcd(double lhs, double rhs, double dTol/* = swMathConst::EPS*/)
//  GCD  :  Greatest Common Divisor
//  use Euclidean Algorithm
//  if return value is one(1), GCD does not exist;
{
	lhs = swMathExt::round(lhs);
	rhs = swMathExt::round(rhs);

	while (1) {
		//lhs = fmod(lhs, rhs);
		lhs = swMathExt::round(fmod(lhs, rhs));
		if (-dTol <= lhs && lhs <= dTol) return rhs;
		//rhs = fmod(rhs, lhs);
		rhs = swMathExt::round(fmod(rhs, lhs));
		if (-dTol <= rhs && rhs <= dTol) return lhs;
	}
}

unsigned long swMathExt::lcm(unsigned long lhs, unsigned long rhs)
//  LCM  :  Least Common Multiple
{  return lhs * rhs / gcd(lhs, rhs);  }

double swMathExt::lcm(double lhs, double rhs, double dTol/* = swMathConst::EPS*/)
//  LCM  :  Least Common Multiple
{  return lhs * rhs / gcd(lhs, rhs, dTol);  }

unsigned long swMathExt::factorial(unsigned long n)
{
	unsigned long ulFactorial = 1ul;
	
	if (n == 0ul || n == 1ul) return 1ul;
	for ( ; n>1ul ; --n) {
		if (std::numeric_limits<unsigned long>::max() / ulFactorial < n) {
			assert(false);
			return ulFactorial;
		}
		ulFactorial *= n;
	}
	return ulFactorial;
}

double swMathExt::factorial(double n, double dTol/* = swMathConst::EPS*/)
{
	double dFactorial = 1.0;
	
	if ((-dTol <= n && n <= dTol) || (-dTol <= n-1.0 && n-1.0 <= dTol)) return 1.0;
	for ( ; n>1ul ; --n) {
		if (std::numeric_limits<double>::max() / dFactorial < n) {
			assert(false);
			return dFactorial;
		}
		dFactorial *= n;
	}
	return dFactorial;
}

unsigned long swMathExt::permutation(unsigned long lhs, unsigned long rhs)
{
	if (lhs < rhs) {
		assert(false);
		return 1ul;
	}
	return factorial(lhs) / factorial(lhs - rhs); 
}

double swMathExt::permutation(double lhs, double rhs, double dTol/* = swMathConst::EPS*/)
{
	if (lhs < rhs) {
		assert(false);
		return 1ul;
	}
	return factorial(lhs, dTol) / factorial(lhs - rhs, dTol); 
}

unsigned long swMathExt::binomial(unsigned long lhs, unsigned long rhs)
{
	if (lhs < rhs) {
		assert(false);
		return 1ul;
	}
	return factorial(lhs) / (factorial(lhs-rhs) * factorial(rhs));
}

//  combination or binomial
double swMathExt::binomial(double lhs, double rhs, double dTol/* = swMathConst::EPS*/)
{
	if (lhs < rhs) {
		assert(false);
		return 1ul;
	}
	return factorial(lhs, dTol) / (factorial(lhs-rhs, dTol) * factorial(rhs, dTol));
}
